#!/usr/bin/env python3
# coding=utf-8

import argparse
from pathlib import Path
import re
import shutil
from typing import (
    List,
    Set,
    Tuple,
)

from romt import base
from romt import common
from romt import dist
from romt import error
from romt import integrity
from romt import signature
from romt.manifest import Manifest, Package

description = """\
Mirror and manage toolchain artifacts.
"""

epilog = """\

SPEC is one of:
  <channel>
  <channel>-YYYY-MM-DD
  <channel>-latest
  <channel>-*
  *-YYYY-MM-DD
  *-latest
  *-*
  YYYY-MM-DD
  latest
  *
Where:
- <channel> is one of: nightly, beta, stable, X.Y.Z
- Single ``*`` represents date, not channel; equivalent to ``*-*``.
- Multiple SPEC options may be given.
- Each SPEC option will be split at commas and whitespace.

TARGET is standard 3- or 4-tuple; common examples:
- x86_64-unknown-linux-gnu (alias ``linux``)
- x86_64-pc-windows-msvc (alias ``windows``)
- x86_64-apple-darwin (alias ``darwin``)

COMMAND values:
  Typical:
    download        download artifacts matching SPEC and TARGET to DEST
    pack            pack DEST artifacts matching SPEC and TARGET into ARCHIVE
    unpack          unpack ARCHIVE into DEST, setting SPEC and TARGET
  Less common:
    fetch-manifest  download manifest matching SPEC and TARGET to DEST
    verify          verify DEST artifacts matching SPEC and TARGET
    list            print DEST artifacts and targets matching SPEC
    all-targets     print all known targets mentioned in SPEC
    fixup           publish manifest variants for SPEC, updating undated

When multiple COMMANDs are given, they share all option values.

For complete details, try ``romt --readme`` to view README.rst.
"""


def parse_spec(spec: str) -> Tuple[str, str]:
    """parse spec into (date, channel).

    Forms with channel:
      <channel>
      <channel>-YYYY-MM-DD
      <channel>-latest
      <channel>-*
      *-YYYY-MM-DD
      *-latest
      *-*
    Forms with only date:
      YYYY-MM-DD
      latest
      *

    Where:
    - <channel> is one of: nightly, beta, stable, X.Y.Z
    - Single ``*`` represents date, not channel; equivalent to ``*-*``.

    """

    # Single "*" is treated as wildcard for date, not channel.
    if spec == "*":
        return "*", "*"

    channel_rex = r"""
        (?P<channel>
            nightly | beta | stable | \* | (?: \d+\.\d+\.\d+ )
        )
        """

    date_rex = r"""
        (?P<date>
            \d\d\d\d-\d\d-\d\d | latest | \*
        )
        """

    m = re.match(
        r"{} (?: - {})? $".format(channel_rex, date_rex), spec, re.VERBOSE
    )
    if m:
        channel = m.group("channel")
        date = m.group("date") or ""
        return date, channel

    m = re.match(r"{} $".format(date_rex), spec, re.VERBOSE)
    if m:
        date = m.group("date")
        return date, "*"

    raise error.UsageError("invalid SPEC {}".format(repr(spec)))


def channel_rel_path(date: str, channel: str) -> str:
    channel_name = "channel-rust-{}.toml".format(channel)
    if date:
        rel_path = "{}/{}".format(date, channel_name)
    else:
        rel_path = "{}".format(channel_name)
    return rel_path


def add_arguments(parser: argparse.ArgumentParser) -> None:
    base.add_downloader_arguments(parser)

    parser.add_argument(
        "-s",
        "--select",
        action="append",
        dest="specs",
        metavar="SPEC",
        default=[],
        help="""one or more SPEC values for toolchain selection""",
    )

    parser.add_argument(
        "-t",
        "--target",
        action="append",
        dest="targets",
        default=[],
        help="""target to download (default varies by COMMAND)""",
    )

    parser.add_argument(
        "--dest",
        action="store",
        default="dist",
        help="""local download directory (default: %(default)s)""",
    )

    parser.add_argument(
        "--url",
        action="store",
        default="https://static.rust-lang.org/dist",
        help="""base URL of dist (default: %(default)s)""",
    )

    parser.add_argument(
        "--archive",
        action="store",
        metavar="ARCHIVE",
        default="toolchain.tar.gz",
        help="use archive ARCHIVE for pack/unpack (default: %(default)s)",
    )

    parser.add_argument(
        "--warn-signature",
        action="store_true",
        default=False,
        help="warn (instead of fail) on signature verification failure",
    )

    parser.add_argument(
        "--no-signature",
        action="store_true",
        default=False,
        help="disable all uses of signature files (*.asc); mainly for testing",
    )

    parser.add_argument(
        "commands",
        nargs="*",
        metavar="COMMAND",
        help="""commands to execute in the order given""",
    )


class Main(dist.DistMain):
    def __init__(self, args: argparse.Namespace) -> None:
        super().__init__(args)
        self.downloader.set_warn_signature(args.warn_signature)
        self._with_sig = not args.no_signature

    def manifest_url_path(self, date: str, channel: str) -> Tuple[str, Path]:
        rel_path = channel_rel_path(date, channel)
        manifest_url = self.url_from_rel_path(rel_path)
        manifest_path = self.dest_path_from_rel_path(rel_path)
        return manifest_url, manifest_path

    def manifest_path(self, date: str, channel: str) -> Path:
        _, man_path = self.manifest_url_path(date, channel)
        return man_path

    def get_manifest(
        self, date: str, channel: str, *, download: bool
    ) -> Manifest:
        man_url, man_path = self.manifest_url_path(date, channel)
        if download:
            if date:
                # Dated manifests may always be cached.
                cached = True
            else:
                # Undated manifests should always be re-downloaded.  They
                # might be newer than what's on disk (either because of the
                # passage of time or because a fixup operation might overwrite
                # the undated manifest with old information before a download).
                cached = False
            self.downloader.download_verify(
                man_url, man_path, cached=cached, with_sig=self._with_sig
            )
        else:
            self.downloader.verify(man_path, with_sig=self._with_sig)
        return Manifest.from_toml_path(man_path)

    def select_manifest(
        self, spec: str, *, download: bool, canonical: bool = False
    ) -> Manifest:
        date, channel = parse_spec(spec)
        manifest = self.get_manifest(date, channel, download=download)
        if canonical and (
            manifest.date != date or manifest.channel != channel
        ):
            manifest = self.get_manifest(
                manifest.date, manifest.channel, download=download
            )
        return manifest

    def channels_in_dest_date(self, date: str) -> List[str]:
        date_path = self.dest_path / date
        prefix = "channel-rust-"
        suffix = ".toml"
        channels = [
            p.name[len(prefix) : -len(suffix)]
            for p in date_path.glob(prefix + "*" + suffix)
        ]
        return channels

    def adjust_download_specs(self, specs: List[str]) -> List[str]:
        # For downloads, require explicit date and channel.
        for spec in specs:
            date, channel = parse_spec(spec)
            if "*" in (date, channel) or date == "latest":
                raise error.UsageError("invalid wild SPEC: {}".format(spec))
        return dist.require_specs(specs)

    def expand_wild_spec(self, spec: str) -> List[str]:
        specs = []  # type: List[str]
        date, channel = parse_spec(spec)
        if "*" in (date, channel) or date == "latest":
            if channel == "*":
                channel_patterns = set("nightly beta stable".split())
            else:
                channel_patterns = set([channel])
            if date in ("*", "latest"):
                dates = common.reversed_date_dir_names(self.dest_path)
            else:
                dates = [date]
            for d in dates:
                channels = channel_patterns.intersection(
                    self.channels_in_dest_date(d)
                )
                specs.extend(
                    "{}-{}".format(channel, d) for channel in channels
                )
                if date == "latest" and specs:
                    break
        else:
            specs.append(spec)

        if not specs:
            raise error.UsageError(
                "no matches for wild SPEC {}".format(repr(spec))
            )

        return specs

    def adjust_wild_specs(self, specs: List[str]) -> List[str]:
        # For non-downloads, handle wild specs.
        adjusted_specs = []
        for spec in specs:
            adjusted_specs.extend(self.expand_wild_spec(spec))
        return dist.require_specs(adjusted_specs)

    def downloaded_packages(self, manifest: Manifest) -> List[Package]:
        packages = []
        for package_info in manifest.gen_available_packages():
            dest_path = self.dest_path_from_rel_path(package_info.rel_path)
            if dest_path.is_file():
                packages.append(package_info)
        return packages

    def downloaded_targets(self, manifest: Manifest) -> List[str]:
        targets = set(p.target for p in self.downloaded_packages(manifest))
        targets.discard("*")
        return sorted(targets)

    def adjust_targets(
        self, manifest: Manifest, base_targets: List[str]
    ) -> List[str]:
        possible_targets = set(p.target for p in manifest.gen_packages())
        targets = set()
        for target in base_targets:
            if target == "all":
                targets.update(possible_targets)
            elif target == "*":
                targets.update(self.downloaded_targets(manifest))
            elif target not in possible_targets:
                raise error.UsageError(
                    "target {} not found in manifest".format(repr(target))
                )
            else:
                targets.add(target)
        return sorted(targets)

    def cmd_fetch_manifest(self) -> None:
        for spec in self.adjust_download_specs(self.specs):
            common.iprint("Fetch manifest: {}".format(spec))
            manifest = self.select_manifest(spec, download=True)
            common.iprint("  ident: {}".format(manifest.ident))

    def _download_verify(
        self, download: bool, specs: List[str], base_targets: List[str]
    ) -> None:
        for spec in specs:
            common.iprint(
                "{}: {}".format("Download" if download else "Verify", spec)
            )
            manifest = self.select_manifest(
                spec, download=download, canonical=True
            )
            common.iprint("  ident: {}".format(manifest.ident))

            targets = self.adjust_targets(manifest, base_targets)
            packages = list(manifest.gen_available_packages(targets=targets))
            common.iprint(
                "  packages: {}, targets: {}".format(
                    len(packages), len(targets)
                )
            )
            for t in targets:
                common.vvprint("  target: {}".format(t))

            for package in packages:
                rel_path = package.rel_path
                dest_path = self.dest_path_from_rel_path(rel_path)
                dest_url = self.url_from_rel_path(rel_path)
                if download:
                    self.downloader.download_verify(
                        dest_url,
                        dest_path,
                        assume_ok=self.args.assume_ok,
                        with_sig=self._with_sig,
                    )
                else:
                    self.downloader.verify(dest_path, with_sig=self._with_sig)

    def cmd_download(self) -> None:
        specs = self.adjust_download_specs(self.specs)
        base_targets = dist.require_targets(self.targets)
        self._download_verify(
            download=True, specs=specs, base_targets=base_targets
        )

    def cmd_verify(self) -> None:
        specs = self.adjust_wild_specs(self.specs)
        base_targets = dist.require_targets(self.targets, default="*")
        self._download_verify(
            download=False, specs=specs, base_targets=base_targets
        )

    def cmd_list(self) -> None:
        max_verbosity = common.get_max_verbosity()
        show_details = max_verbosity >= common.VERBOSITY_INFO
        for spec in self.adjust_wild_specs(self.specs):
            common.vprint("List: {}".format(spec))
            manifest = self.select_manifest(spec, download=False)
            if show_details:
                available_packages = manifest.available_packages()
                available_targets = manifest.available_targets()
                packages = self.downloaded_packages(manifest)
                targets = self.downloaded_targets(manifest)

                target_out = "targets[{}/{}]".format(
                    len(targets), len(available_targets),
                )
                package_out = "packages[{}/{}]".format(
                    len(packages), len(available_packages),
                )
                # Example output:
                #   stable-2020-01-30(1.41.0)    \
                #     targets[84/84], packages[272/326]
                common.iprint(
                    "{:28} {:16} {:18}".format(
                        manifest.ident, target_out, package_out
                    )
                )
                for target in targets:
                    common.iprint("  {}".format(target))
            else:
                common.eprint(manifest.ident)

    def cmd_all_targets(self) -> None:
        for spec in self.adjust_wild_specs(self.specs):
            common.iprint("All targets: {}".format(spec))
            manifest = self.select_manifest(spec, download=False)
            common.iprint("  ident: {}".format(manifest.ident))
            for target in manifest.all_targets():
                common.eprint(target)

    def cmd_pack(self) -> None:
        base_targets = dist.require_targets(self.targets, default="*")
        archive_path = self.get_archive_path()
        common.iprint("Packing archive: {}".format(archive_path))
        with common.tar_context(archive_path, "w") as tar_f:

            def pack_path(rel_path: str) -> None:
                dest_path = self.dest_path_from_rel_path(rel_path)
                packed_name = "dist/" + rel_path
                common.vprint("[pack] {}".format(rel_path))
                try:
                    tar_f.add(str(dest_path), packed_name)
                except FileNotFoundError:
                    raise error.MissingFileError(str(dest_path))

            def pack_rel_path(rel_path: str) -> None:
                pack_path(rel_path)
                pack_path(integrity.append_hash_suffix(rel_path))
                if self._with_sig:
                    pack_path(signature.append_sig_suffix(rel_path))

            for spec in self.adjust_wild_specs(self.specs):
                common.iprint("Pack: {}".format(spec))
                manifest = self.select_manifest(
                    spec, download=False, canonical=True
                )
                common.iprint("  ident: {}".format(manifest.ident))

                targets = self.adjust_targets(manifest, base_targets)
                packages = list(
                    manifest.gen_available_packages(targets=targets)
                )
                common.iprint(
                    "  packages: {}, targets: {}".format(
                        len(packages), len(targets)
                    )
                )
                for t in targets:
                    common.vvprint("  target: {}".format(t))

                # Pack channel file.
                pack_rel_path(
                    channel_rel_path(manifest.date, manifest.channel)
                )

                # Pack up package file parts.
                for package in packages:
                    pack_rel_path(package.rel_path)

    def _detect_specs(self, rel_paths: Set[str]) -> List[str]:
        specs = []
        for rel_path in rel_paths:
            m = re.match(
                r"""
                (?P<date>\d\d\d\d-\d\d-\d\d)
                /channel-rust-
                (?P<channel>nightly|stable|beta)
                \.toml$
                """,
                rel_path,
                re.VERBOSE,
            )
            if m:
                date = m.group("date")
                channel = m.group("channel")
                specs.append("{}-{}".format(channel, date))
        return specs

    def _detect_targets(
        self, specs: List[str], rel_paths: Set[str]
    ) -> List[str]:
        targets = set()
        for spec in specs:
            manifest = self.select_manifest(spec, download=False)
            for package in manifest.gen_available_packages():
                if package.target != "*" and package.rel_path in rel_paths:
                    targets.add(package.target)
        return sorted(targets)

    def cmd_unpack(self) -> None:
        archive_path = self.get_archive_path()
        common.iprint("Unpacking archive: {}".format(archive_path))
        dist_prefix = "dist/"
        extracted = set()
        with common.tar_context(archive_path, "r") as tar_f:
            for tar_info in tar_f:
                if tar_info.isdir():
                    continue
                if not tar_info.name.startswith(dist_prefix):
                    raise error.UnexpectedArchiveMemberError(tar_info.name)

                rel_path = tar_info.name[len(dist_prefix) :]
                dest_path = self.dest_path_from_rel_path(rel_path)
                tar_info.name = str(dest_path)
                common.vprint("[unpack] {}".format(rel_path))
                tar_f.extract(tar_info)
                extracted.add(rel_path)

        specs = self._detect_specs(extracted)
        targets = self._detect_targets(specs, extracted)

        common.iprint("Unpacked specs: {}".format(len(specs)))
        for spec in specs:
            common.iprint("  {}".format(spec))

        common.iprint("Unpacked targets: {}".format(len(targets)))
        for target in targets:
            common.iprint("  {}".format(target))

        self.specs = specs
        self.targets = targets

    def _write_manifest(
        self, manifest: Manifest, date: str, channel: str
    ) -> None:
        src_path = self.manifest_path(manifest.date, manifest.channel)
        dst_path = self.manifest_path(date, channel)
        common.iprint("[publish] {}".format(dst_path))
        shutil.copyfile(src_path, dst_path)
        src_hash_path = integrity.path_append_hash_suffix(src_path)
        dst_hash_path = integrity.path_append_hash_suffix(dst_path)
        shutil.copyfile(src_hash_path, dst_hash_path)
        src_sig_path = signature.path_append_sig_suffix(src_path)
        dst_sig_path = signature.path_append_sig_suffix(dst_path)
        shutil.copyfile(src_sig_path, dst_sig_path)

    def _write_manifest_variations(self, manifest: Manifest) -> None:
        date = manifest.date
        channel = manifest.channel

        # Write top-level manifest, unless a newer one already exists, e.g.:
        #   dist/channel-rust-nightly.toml
        top_path = self.manifest_path("", channel)
        if top_path.is_file():
            top_man = self.get_manifest("", channel, download=False)
            write_top = manifest.date >= top_man.date
        else:
            write_top = True
        if write_top:
            self._write_manifest(manifest, "", channel)

        if channel == "stable":
            # Write version-numbered manifests, e.g.:
            #   dist/yyyy-mm-dd/channel-rust-x.y.z.toml
            #   dist/channel-rust-x.y.z.toml
            self._write_manifest(manifest, date, manifest.version)
            self._write_manifest(manifest, "", manifest.version)

    def cmd_fixup(self) -> None:
        for spec in self.adjust_wild_specs(self.specs):
            common.iprint("Fixup: {}".format(spec))
            manifest = self.select_manifest(
                spec, download=False, canonical=True
            )
            common.vprint("  ident: {}".format(manifest.ident))
            self._write_manifest_variations(manifest)

    def run(self) -> None:
        valid_commands = [
            "fetch-manifest",
            "download",
            "verify",
            "list",
            "all-targets",
            "pack",
            "unpack",
            "fixup",
        ]
        commands = self.args.commands
        if commands:
            base.verify_commands(commands, valid_commands)
        else:
            common.iprint("nothing to do; try a COMMAND")
            return

        for cmd in commands:
            if cmd == "fetch-manifest":
                self.cmd_fetch_manifest()

            elif cmd == "download":
                self.cmd_download()
                self.cmd_fixup()

            elif cmd == "verify":
                self.cmd_verify()

            elif cmd == "list":
                self.cmd_list()

            elif cmd == "all-targets":
                self.cmd_all_targets()

            elif cmd == "pack":
                self.cmd_pack()

            elif cmd == "unpack":
                self.cmd_unpack()
                self.cmd_verify()
                self.cmd_fixup()

            elif cmd == "fixup":
                self.cmd_fixup()



#!/usr/bin/env python3
# coding=utf-8

import argparse
from pathlib import Path
import shutil
from typing import (
    List,
    Set,
    Tuple,
)

import toml

from romt import base
from romt import common
from romt import dist
from romt import error
from romt import integrity

description = """\
Mirror and manage rustup tooling.
"""

epilog = """\

SPEC is one of:
  <version>
  stable
  latest
  *
Where:
- <version> is: X.Y.Z
- Multiple SPEC options may be given.
- Each SPEC option will be split at commas and whitespace.

TARGET is standard 3- or 4-tuple; common examples:
- x86_64-unknown-linux-gnu (alias ``linux``)
- x86_64-pc-windows-msvc (alias ``windows``)
- x86_64-apple-darwin (alias ``darwin``)

COMMAND values:

  download          download artifacts matching SPEC and TARGET to DEST
  verify            verify DEST artifacts matching SPEC and TARGET
  list              print DEST artifacts and targets matching SPEC
  all-targets       print all known targets (hard-coded; others are permitted)
  pack              pack DEST artifacts matching SPEC and TARGET into ARCHIVE
  unpack            unpack ARCHIVE into DEST, setting SPEC and TARGET
  fixup             fixup latest stable release with SPEC as candidate

When multiple COMMANDs are given, they share all option values.

For complete details, try ``romt --readme`` to view README.rst.
"""

# Note: Known targets are found by inspecting the S3 tree, e.g.::
#  aws s3 ls --no-sign-request s3://static-rust-lang-org/rustup/archive/1.21.1/

ALL_KNOWN_TARGETS = """
    aarch64-linux-android
    aarch64-unknown-linux-gnu
    arm-linux-androideabi
    arm-unknown-linux-gnueabi
    arm-unknown-linux-gnueabihf
    armv7-linux-androideabi
    armv7-unknown-linux-gnueabihf
    i686-apple-darwin
    i686-linux-android
    i686-pc-windows-gnu
    i686-pc-windows-msvc
    i686-unknown-linux-gnu
    mips-unknown-linux-gnu
    mips64-unknown-linux-gnuabi64
    mips64el-unknown-linux-gnuabi64
    mipsel-unknown-linux-gnu
    powerpc-unknown-linux-gnu
    powerpc64-unknown-linux-gnu
    powerpc64le-unknown-linux-gnu
    s390x-unknown-linux-gnu
    x86_64-apple-darwin
    x86_64-linux-android
    x86_64-pc-windows-gnu
    x86_64-pc-windows-msvc
    x86_64-unknown-freebsd
    x86_64-unknown-linux-gnu
    x86_64-unknown-linux-musl
    x86_64-unknown-netbsd
    """.split()


def validate_spec(spec: str) -> str:
    """parse spec into (date, channel).

    SPEC is one of:
      <version>
      stable
      latest
      *
    Where:
    - <version> is: X.Y.Z
    - Multiple SPEC options may be given.
    - Each SPEC option will be split at commas and whitespace.

    """

    if spec in ("*", "latest", "stable") or common.is_version(spec):
        return spec

    raise error.UsageError("invalid SPEC {}".format(repr(spec)))


def add_arguments(parser: argparse.ArgumentParser) -> None:
    base.add_downloader_arguments(parser)

    parser.add_argument(
        "-s",
        "--select",
        action="append",
        dest="specs",
        metavar="SPEC",
        default=[],
        help="""one or more SPEC values for rustup selection""",
    )

    parser.add_argument(
        "-t",
        "--target",
        action="append",
        dest="targets",
        default=[],
        help="""target to download (default varies by COMMAND)""",
    )

    parser.add_argument(
        "--dest",
        action="store",
        default="rustup",
        help="""local download directory (default: %(default)s)""",
    )

    parser.add_argument(
        "--url",
        action="store",
        default="https://static.rust-lang.org/rustup",
        help="""base URL of rustup (default: %(default)s)""",
    )

    parser.add_argument(
        "--archive",
        action="store",
        metavar="ARCHIVE",
        default="rustup.tar.gz",
        help="use archive ARCHIVE for pack/unpack (default: %(default)s)",
    )

    parser.add_argument(
        "commands",
        nargs="*",
        metavar="COMMAND",
        help="""commands to execute in the order given""",
    )


class Main(dist.DistMain):
    def __init__(self, args: argparse.Namespace) -> None:
        super().__init__(args)

    @property
    def release_stable_url_path(self) -> Tuple[str, Path]:
        rel_path = "release-stable.toml"
        release_stable_url = self.url_from_rel_path(rel_path)
        release_stable_path = self.dest_path_from_rel_path(rel_path)
        return release_stable_url, release_stable_path

    @property
    def release_stable_path(self) -> Path:
        _, path = self.release_stable_url_path
        return path

    def get_release_stable_version(self, *, download: bool) -> str:
        url, path = self.release_stable_url_path
        if download:
            # This file changes unexpectedly.  Avoid caching to ensure the
            # correct version is used.
            self.downloader.download_cached(url, path, cached=False)
        elif path.is_file():
            common.vprint("[read] {}".format(path))
        else:
            raise error.MissingFileError(str(path))
        return toml.load(path)["version"]

    @property
    def artifact_root_rel_path(self) -> str:
        return "archive"

    @property
    def artifact_root_path(self) -> Path:
        return self.dest_path_from_rel_path(self.artifact_root_rel_path)

    def artifact_version_rel_path(self, version: str) -> str:
        return "{}/{}".format(self.artifact_root_rel_path, version)

    def artifact_version_path(self, version: str) -> Path:
        return self.dest_path_from_rel_path(
            self.artifact_version_rel_path(version)
        )

    def rustup_init_rel_path(self, version: str, target: str) -> str:
        # archive/<version>/<target>/rustup-init[.exe]
        base_rel_path = "{}/{}/rustup-init".format(
            self.artifact_version_rel_path(version), target
        )
        rel_path = dist.append_exe_suffix(base_rel_path, target)
        return rel_path

    def version_from_spec(self, spec: str, *, download: bool) -> str:
        if spec == "stable":
            return self.get_release_stable_version(download=download)
        else:
            return spec

    def downloaded_versions(self) -> List[str]:
        versions = [p.name for p in self.artifact_root_path.glob("*")]
        return common.reverse_sorted_versions(versions)

    def adjust_download_specs(self, specs: List[str]) -> List[str]:
        # For downloads, wildcards not permitted.
        for spec in specs:
            validate_spec(spec)
            if spec == "*" or spec == "latest":
                raise error.UsageError("invalid wild SPEC: {}".format(spec))
        return dist.require_specs(specs)

    def expand_wild_spec(self, spec: str) -> List[str]:
        specs = []  # type: List[str]
        validate_spec(spec)
        if spec == "*":
            specs.extend(self.downloaded_versions())

        elif spec == "latest":
            found_versions = self.downloaded_versions()
            if found_versions:
                specs.append(found_versions[0])

        else:
            specs.append(spec)

        if not specs:
            raise error.UsageError(
                "no matches for wild SPEC {}".format(repr(spec))
            )

        return specs

    def adjust_wild_specs(self, specs: List[str]) -> List[str]:
        # For non-downloads, handle wild specs.
        adjusted_specs = []
        for spec in specs:
            adjusted_specs.extend(self.expand_wild_spec(spec))
        return dist.require_specs(adjusted_specs)

    def downloaded_targets(self, version: str) -> List[str]:
        target_paths = common.gen_dirs(self.artifact_version_path(version))
        return sorted(t.name for t in target_paths)

    def adjust_targets(
        self, version: str, base_targets: List[str]
    ) -> List[str]:
        targets = set()
        known_targets = set(ALL_KNOWN_TARGETS)
        for target in base_targets:
            if target == "all":
                targets.update(known_targets)
            elif target == "*":
                targets.update(self.downloaded_targets(version))
            else:
                if target not in known_targets:
                    common.eprint(
                        "warning: unknown target {}".format(repr(target))
                    )
                targets.add(target)
        return sorted(targets)

    def _download_verify(
        self, download: bool, specs: List[str], base_targets: List[str]
    ) -> None:
        for spec in specs:
            common.iprint(
                "{}: {}".format("Download" if download else "Verify", spec)
            )
            version = self.version_from_spec(spec, download=download)
            common.iprint("  version: {}".format(version))

            targets = self.adjust_targets(version, base_targets)
            common.iprint("  targets: {}".format(len(targets)))
            for t in targets:
                common.vvprint("  target: {}".format(t))

            for target in targets:
                rel_path = self.rustup_init_rel_path(version, target)
                dest_path = self.dest_path_from_rel_path(rel_path)
                dest_url = self.url_from_rel_path(rel_path)
                if download:
                    self.downloader.download_verify(
                        dest_url, dest_path, assume_ok=self.args.assume_ok
                    )
                else:
                    self.downloader.verify(dest_path)

    def cmd_download(self) -> None:
        specs = self.adjust_download_specs(self.specs)
        base_targets = dist.require_targets(self.targets)
        self._download_verify(
            download=True, specs=specs, base_targets=base_targets
        )

    def cmd_verify(self) -> None:
        specs = self.adjust_wild_specs(self.specs)
        base_targets = dist.require_targets(self.targets, default="*")
        self._download_verify(
            download=False, specs=specs, base_targets=base_targets
        )

    def cmd_list(self) -> None:
        max_verbosity = common.get_max_verbosity()
        show_details = max_verbosity >= common.VERBOSITY_INFO
        for spec in self.adjust_wild_specs(self.specs):
            common.iprint("List: {}".format(spec))
            version = self.version_from_spec(spec, download=False)
            if show_details:
                targets = self.downloaded_targets(version)

                target_out = "targets[{}]".format(len(targets))
                # Example output:
                #   1.41.0   targets[84]
                common.iprint("{:8} {}".format(version, target_out))
                for target in targets:
                    common.iprint("  {}".format(target))
            else:
                common.eprint(version)

    def cmd_all_targets(self) -> None:
        common.iprint("All known targets:")
        for target in ALL_KNOWN_TARGETS:
            common.eprint(target)

    def cmd_pack(self) -> None:
        base_targets = dist.require_targets(self.targets, default="*")
        archive_path = self.get_archive_path()
        common.iprint("Packing archive: {}".format(archive_path))
        with common.tar_context(archive_path, "w") as tar_f:

            def pack_path(rel_path: str) -> None:
                dest_path = self.dest_path_from_rel_path(rel_path)
                packed_name = "rustup/" + rel_path
                common.vprint("[pack] {}".format(rel_path))
                try:
                    tar_f.add(str(dest_path), packed_name)
                except FileNotFoundError:
                    raise error.MissingFileError(str(dest_path))

            def pack_rel_path(rel_path: str) -> None:
                pack_path(rel_path)
                pack_path(integrity.append_hash_suffix(rel_path))

            for spec in self.adjust_wild_specs(self.specs):
                common.iprint("Pack: {}".format(spec))
                version = self.version_from_spec(spec, download=False)
                common.iprint("  version: {}".format(version))

                targets = self.adjust_targets(version, base_targets)
                common.iprint("  targets: {}".format(len(targets)))
                for t in targets:
                    common.vvprint("  target: {}".format(t))

                for target in targets:
                    rel_path = self.rustup_init_rel_path(version, target)
                    pack_rel_path(rel_path)

    def _detect_version_targets(
        self, rel_paths: Set[str]
    ) -> Tuple[List[str], List[str]]:
        # rel_paths should be: "archive/<version>/<target>/<file>".
        versions = set()
        targets = set()
        for p in rel_paths:
            parts = p.split("/")
            if len(parts) < 4:
                common.eprint("warning: unexpected path {}".format(p))
            else:
                version = parts[1]
                target = parts[2]
                if common.is_version(version):
                    versions.add(version)
                    targets.add(target)
        return sorted(versions), sorted(targets)

    def cmd_unpack(self) -> None:
        archive_path = self.get_archive_path()
        common.iprint("Unpacking archive: {}".format(archive_path))
        rustup_prefix = "rustup/"
        prefix = "{}{}/".format(rustup_prefix, self.artifact_root_rel_path)
        extracted = set()
        with common.tar_context(archive_path, "r") as tar_f:
            for tar_info in tar_f:
                if tar_info.isdir():
                    continue
                if not tar_info.name.startswith(prefix):
                    raise error.UnexpectedArchiveMemberError(tar_info.name)

                rel_path = tar_info.name[len(rustup_prefix) :]
                dest_path = self.dest_path_from_rel_path(rel_path)
                tar_info.name = str(dest_path)
                common.vprint("[unpack] {}".format(rel_path))
                tar_f.extract(tar_info)
                extracted.add(rel_path)

        specs, targets = self._detect_version_targets(extracted)

        common.iprint("Unpacked specs: {}".format(len(specs)))
        for spec in specs:
            common.iprint("  {}".format(spec))

        common.iprint("Unpacked targets: {}".format(len(targets)))
        for target in targets:
            common.iprint("  {}".format(target))

        self.specs = specs
        self.targets = targets

    def _write_release_stable(self, version: str) -> None:
        path = self.release_stable_path
        toml_dict = {"schema-version": "1", "version": version}
        with path.open("w") as f:
            toml.dump(toml_dict, f)

    def _fixup_version(self, version: str) -> None:
        # Write release_stable unless a newer one already exists.
        path = self.release_stable_path
        if path.is_file():
            old_version = self.get_release_stable_version(download=False)
            new_key = common.version_sort_key(version)
            old_key = common.version_sort_key(old_version)
            write = new_key >= old_key
        else:
            write = True
        if write:
            common.iprint("[write] {} (version={})".format(path, version))
            self._write_release_stable(version)

    def cmd_fixup(self) -> None:
        for spec in self.adjust_wild_specs(self.specs):
            common.iprint("Fixup: {}".format(spec))
            version = self.version_from_spec(spec, download=False)
            version_path = self.artifact_version_path(version)
            # Artifacts are arranged as: <version_path>/<target>/<artifact>
            # If no artifacts exist for any targets, claim the version
            # is not present.
            artifacts = list(version_path.glob("*/*"))
            if not artifacts:
                raise error.UsageError(
                    "version {} not present".format(version)
                )
            self._fixup_version(version)

        # Copy rustup/archive/<stable_version>/ to rustup/dist/.
        stable_version = self.get_release_stable_version(download=False)

        archive_version_path = self.dest_path_from_rel_path(
            "archive/{}".format(stable_version)
        )
        if not archive_version_path.is_dir():
            raise error.MissingDirectoryError(str(archive_version_path))

        dist_path = self.dest_path_from_rel_path("dist")
        if dist_path.is_dir():
            shutil.rmtree(dist_path)

        common.iprint(
            "[copytree] {} -> {}".format(archive_version_path, dist_path)
        )
        shutil.copytree(archive_version_path, dist_path)

    def run(self) -> None:
        valid_commands = [
            "download",
            "verify",
            "list",
            "all-targets",
            "pack",
            "unpack",
            "fixup",
        ]
        commands = self.args.commands
        if commands:
            base.verify_commands(commands, valid_commands)
        else:
            common.iprint("nothing to do; try a COMMAND")

        for cmd in commands:
            if cmd == "download":
                self.cmd_download()
                self.cmd_fixup()

            elif cmd == "verify":
                self.cmd_verify()

            elif cmd == "list":
                self.cmd_list()

            elif cmd == "all-targets":
                self.cmd_all_targets()

            elif cmd == "pack":
                self.cmd_pack()

            elif cmd == "unpack":
                self.cmd_unpack()
                self.cmd_verify()
                self.cmd_fixup()

            elif cmd == "fixup":
                self.cmd_fixup()



#!/usr/bin/env python3
# coding=utf-8

import pytest

from romt import error
import romt.toolchain


def test_parse_select() -> None:
    parse = romt.toolchain.parse_spec
    # Channel-only; date is "":
    assert parse("nightly") == ("", "nightly")
    assert parse("stable") == ("", "stable")
    assert parse("beta") == ("", "beta")
    assert parse("1.23.0") == ("", "1.23.0")
    # Channel and date:
    assert parse("nightly-2020-04-01") == ("2020-04-01", "nightly")
    assert parse("stable-latest") == ("latest", "stable")
    assert parse("beta-*") == ("*", "beta")
    # Date-only; channel is "*":
    assert parse("2020-04-01") == ("2020-04-01", "*")
    assert parse("latest") == ("latest", "*")
    assert parse("*") == ("*", "*")

    with pytest.raises(error.UsageError):
        parse("nightly-")
    with pytest.raises(error.UsageError):
        parse("-latest")
    with pytest.raises(error.UsageError):
        parse("lateststuff")
    with pytest.raises(error.UsageError):
        parse("**")



#!/usr/bin/env python3
# coding=utf-8

from contextlib import contextmanager
import datetime
import os
from pathlib import Path
import platform
import re
import stat
import tarfile
from typing import Any, Generator, IO, Iterable, List, Optional, Tuple


is_windows = platform.system() == "Windows"


VERBOSITY_ERROR = 0
VERBOSITY_INFO = 1
VERBOSITY_VERBOSE = 2
VERBOSITY_VVERBOSE = 3
_max_verbosity = VERBOSITY_INFO


def get_max_verbosity() -> int:
    return _max_verbosity


def set_max_verbosity(max_verbosity: int) -> None:
    global _max_verbosity
    _max_verbosity = max_verbosity


def _print_verbosity(verbosity: int, *args: Any, **kwargs: Any) -> None:
    if verbosity <= _max_verbosity:
        print(*args, flush=True, **kwargs)


def vvprint(*args: Any, **kwargs: Any) -> None:
    _print_verbosity(VERBOSITY_VVERBOSE, *args, **kwargs)


def vprint(*args: Any, **kwargs: Any) -> None:
    _print_verbosity(VERBOSITY_VERBOSE, *args, **kwargs)


def iprint(*args: Any, **kwargs: Any) -> None:
    _print_verbosity(VERBOSITY_INFO, *args, **kwargs)


def eprint(*args: Any, **kwargs: Any) -> None:
    _print_verbosity(VERBOSITY_ERROR, *args, **kwargs)


def is_date(date: str) -> bool:
    return re.match(r"\d\d\d\d-\d\d-\d\d$", date) is not None


def is_version(version: str) -> bool:
    return re.match(r"\d+\.\d+\.\d+$", version) is not None


def version_sort_key(version: str) -> Tuple[int, ...]:
    return tuple(int(v) for v in version.split("."))


def reverse_sorted_versions(versions: List[str]) -> List[str]:
    return sorted(versions, key=version_sort_key, reverse=True)


def path_append(path: Path, suffix: str) -> Path:
    return path.with_name(path.name + suffix)


def tmp_path_for(path: Path) -> Path:
    return path.with_name("." + path.name + ".tmp")


def path_modified_today(path: Path) -> bool:
    if path.is_file():
        path_mtime = path.stat().st_mtime
        path_mdate = datetime.datetime.fromtimestamp(path_mtime).date()
        today_date = datetime.datetime.now().date()
        modified_today = path_mdate == today_date
    else:
        modified_today = False
    return modified_today


def get_umask() -> int:
    umask = os.umask(0)
    os.umask(umask)
    return umask


def is_executable(path: Path) -> bool:
    return os.access(str(path), os.X_OK)


def chmod_executable(path: Path) -> None:
    x_bits = (stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH) & ~get_umask()
    os.chmod(str(path), path.stat().st_mode | x_bits)


def gen_dirs(parent: Path) -> Generator[Path, None, None]:
    """generate Path for each dir in parent."""
    for candidate in parent.glob("*"):
        if candidate.is_dir():
            yield candidate


def reversed_date_dir_names(parent: Path) -> List[str]:
    """list of yyyy-mm-dd dirnames in parent (newest to oldest)."""
    dirs = sorted(
        (d.name for d in gen_dirs(parent) if is_date(d.name)), reverse=True,
    )
    return dirs


def open_optional(path: str, mode: str) -> Optional[IO[Any]]:
    return open(path, mode) if path else None


def close_optional(f: Optional[IO[Any]]) -> None:
    if f:
        f.close()


def make_dirs_for(path: Path) -> None:
    parent = path.parent
    if not parent.is_dir():
        parent.mkdir(parents=True)


def log(log_file: Optional[IO[Any]], message: Any) -> None:
    if log_file is not None:
        log_file.write("{}\n".format(message))
        log_file.flush()


def split_word(item: str) -> List[str]:
    """split item into list of words at commas or runs of whitespace.

    Retains any duplicates and empty strings.
    """
    return [part for part in re.split(r"(?:,|\s+)", item)]


def split_flatten_words(words: Iterable[str]) -> List[str]:
    """split_word(each_word in words) into flattened, deduped list.
    """
    dedup = set()
    result = []
    for w in words:
        for part in split_word(w):
            if part not in dedup:
                dedup.add(part)
                result.append(part)
    return result


def split_flatten_normalize_words(words: Iterable[str]) -> List[str]:
    """split_flatten_words(), remove dups and empty, sort.
    """
    norm_words = set(w for w in split_flatten_words(words) if w)
    return sorted(norm_words)


def normalize_patterns(patterns: Iterable[str]) -> List[str]:
    """split, flatten, remove dups and empty, reduce "*", sort."""
    norm_patterns = split_flatten_normalize_words(patterns)
    if "*" in norm_patterns:
        return ["*"]
    return norm_patterns


@contextmanager
def tar_context(
    archive_path: Path, mode: str
) -> Generator[tarfile.TarFile, None, None]:
    """mode is "r" (read) or "w" (write)."""
    writing = mode == "w"
    if archive_path.name.endswith(".gz"):
        mode += ":gz"

    if writing:
        if archive_path.exists():
            archive_path.unlink()
        tmp_archive_path = tmp_path_for(archive_path)
        tar_f = tarfile.open(tmp_archive_path, mode)
        try:
            yield tar_f
        except (Exception, KeyboardInterrupt):
            tar_f.close()
            if tmp_archive_path.is_file():
                tmp_archive_path.unlink()
            raise
        tar_f.close()
        tmp_archive_path.rename(archive_path)
    else:
        tar_f = tarfile.open(archive_path, mode)
        yield tar_f
        tar_f.close()



#!/usr/bin/env python3
# coding=utf-8

import hashlib
from pathlib import Path
from typing import BinaryIO, Optional

from romt import common
from romt.error import (
    IntegrityError,
    MissingFileError,
)


def hash_bytes(data: bytes) -> str:
    h = hashlib.sha256()
    h.update(data)
    return h.hexdigest()


def hash_fileobj(fileobj: BinaryIO) -> str:
    h = hashlib.sha256()
    while True:
        chunk = fileobj.read(8192)
        if not chunk:
            break
        h.update(chunk)
    return h.hexdigest()


def hash_file(path: Path) -> str:
    with path.open("rb") as f:
        hash = hash_fileobj(f)
    return hash


def parse_hash_text(hash_text: str) -> str:
    # Expected format: text with lines:
    #   <sha256>  filename
    return hash_text.split()[0]


def read_hash_file(path: Path) -> str:
    return parse_hash_text(path.read_text("utf-8"))


_hash_suffix = ".sha256"


def append_hash_suffix(s: str) -> str:
    return s + _hash_suffix


def path_append_hash_suffix(path: Path) -> Path:
    return common.path_append(path, _hash_suffix)


def write_hash_file_for(path: Path) -> None:
    h = hash_file(path)
    path_sha256 = path_append_hash_suffix(path)
    with path_sha256.open("w") as f:
        f.write("{}  {}\n".format(h, path.name))


def verify_hash(path: Path, expected_hash: str) -> None:
    """
    Raises:
        MissingFileError - path doesn't exist
        IntegrityError - path exists with bad hash
    """

    if not path.exists():
        raise MissingFileError(str(path))
    h = hash_file(path)
    if h != expected_hash:
        raise IntegrityError(
            path.name, actual_hash=h, expected_hash=expected_hash
        )


def verify(path: Path, hash_path: Optional[Path] = None) -> None:
    """
    Raises:
        MissingFileError - either path doesn't exist
        IntegrityError - paths exists with bad hash
    """

    if hash_path is None:
        hash_path = path_append_hash_suffix(path)
    if not hash_path.exists():
        raise MissingFileError(str(hash_path))
    verify_hash(path, read_hash_file(hash_path))



#!/usr/bin/env python3
# coding=utf-8

from romt import common


def test_is_date() -> None:
    is_date = common.is_date
    assert is_date("2020-01-23")
    assert not is_date("2020-01-2")
    assert not is_date("2020-01-234")
    assert not is_date("20-01-23")


def test_is_version() -> None:
    is_version = common.is_version
    assert is_version("1.2.0")
    assert is_version("1.2.3")
    assert not is_version("1.2")
    assert not is_version("a.b.c")


def test_split_word() -> None:
    assert common.split_word("") == [""]
    assert common.split_word("a") == ["a"]
    assert common.split_word("a,b") == ["a", "b"]
    assert common.split_word("a,") == ["a", ""]
    assert common.split_word(",") == ["", ""]
    assert common.split_word("a    b") == ["a", "b"]


def test_split_flatten_words() -> None:
    split_flatten_words = common.split_flatten_words
    assert split_flatten_words([""]) == [""]
    assert split_flatten_words(["a"]) == ["a"]
    assert split_flatten_words(["a", "b"]) == ["a", "b"]
    assert split_flatten_words(["b,a", "b"]) == ["b", "a"]
    assert split_flatten_words([",a", "b    c"]) == ["", "a", "b", "c"]


def test_split_flatten_normalize_words() -> None:
    f = common.split_flatten_normalize_words
    assert f([""]) == []
    assert f(["  c  b,,,a "]) == ["a", "b", "c"]
    assert f(["d,b", "c , a"]) == ["a", "b", "c", "d"]


def test_normalize_patterns() -> None:
    normalize_patterns = common.normalize_patterns
    assert normalize_patterns([""]) == []
    assert normalize_patterns(["  c  b,,,a "]) == ["a", "b", "c"]
    assert normalize_patterns(["d,b", "c , a"]) == ["a", "b", "c", "d"]
    assert normalize_patterns(["a,*,b"]) == ["*"]



#!/usr/bin/env python
# coding=utf-8

import setuptools

NAME = "romt"

description = """
Romt (Rust Offline Mirror Tool) enables mirroring of Rust programming
language tools and crates for use in an offline context.
""".strip().replace("\n", " ")

keywords = "Rust mirror toolchain crates"


__version__ = None
for line in open("src/{}/cli.py".format(NAME), encoding="utf-8"):
    if line.startswith("__version__"):
        __version__ = line.split('"')[1]
        break

with open("README.rst", encoding="utf-8") as f:
    long_description = f.read()

with open("requirements.txt", encoding="utf-8") as f:
    requirements = f.read()

with open("dev-requirements.txt", encoding="utf-8") as f:
    dev_requirements = f.read()

setuptools.setup(
    name=NAME,
    version=__version__,
    packages=setuptools.find_packages("src"),
    package_dir={"": "src"},
    python_requires=">=3.5",
    install_requires=requirements,
    extras_require={"dev": dev_requirements},
    entry_points={"console_scripts": ["{}={}.cli:main".format(NAME, NAME)]},
    description=description,
    long_description=long_description,
    keywords=keywords,
    url="https://github.com/drmikehenry/romt",
    author="Michael Henry",
    author_email="drmikehenry@drmikehenry.com",
    license="MIT",
    zip_safe=True,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Topic :: Utilities",
        "Environment :: Console",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
)



#!/usr/bin/env python3
# coding=utf-8

import io
from pathlib import Path
from typing import BinaryIO

import requests

from romt import common
from romt import error
from romt import integrity
from romt import signature


class Downloader:
    def __init__(self) -> None:
        self._session = requests.Session()
        self.sig_verifier = signature.Verifier()
        self._warn_signature = False

    def set_warn_signature(self, warn_signature: bool) -> None:
        self._warn_signature = warn_signature

    def download_fileobj(self, url: str, fileobj: BinaryIO) -> None:
        try:
            response = self._session.get(url, stream=True)
            response.raise_for_status()
            for chunk in response.iter_content(chunk_size=4096):
                fileobj.write(chunk)
        except requests.exceptions.RequestException as e:
            raise error.DownloadError(url, e)

    def get(self, url: str) -> bytes:
        data = io.BytesIO()
        self.download_fileobj(url, data)
        return data.getvalue()

    def get_text(self, url: str) -> str:
        return self.get(url).decode("utf-8")

    def _download(self, url: str, dest_path: Path) -> None:
        common.make_dirs_for(dest_path)
        try:
            with open(dest_path, "wb") as f:
                self.download_fileobj(url, f)
        except error.DownloadError:
            if dest_path.is_file():
                dest_path.unlink()
            raise

    def download(self, url: str, dest_path: Path) -> None:
        if dest_path.is_file():
            dest_path.unlink()
        tmp_dest_path = common.tmp_path_for(dest_path)
        self._download(url, tmp_dest_path)
        tmp_dest_path.rename(dest_path)

    def download_cached(
        self, dest_url: str, dest_path: Path, *, cached: bool = True
    ) -> None:
        if cached and dest_path.is_file():
            common.vprint("[cached file] {}".format(dest_path))
        else:
            common.vprint("[downloading] {}".format(dest_path))
            self.download(dest_url, dest_path)

    def sig_verify(self, path: Path, sig_path: Path) -> None:
        try:
            self.sig_verifier.verify(path, sig_path)
        except (error.MissingFileError, error.IntegrityError):
            if self._warn_signature:
                common.eprint("Warning: Signature failure for {}".format(path))
            else:
                raise

    def verify_hash(self, path: Path, hash: str) -> None:
        """
        Raises:
            MissingFileError - path doesn't exist
            IntegrityError - path exists with bad hash
        """
        common.vprint("[verify] {}".format(path))
        integrity.verify_hash(path, hash)

    def verify(self, path: Path, *, with_sig: bool = False) -> None:
        """
        Raises:
            MissingFileError - path or associated sha256path doesn't exist
            IntegrityError - paths exists with bad hash
        """
        hash_path = integrity.path_append_hash_suffix(path)
        common.vprint("[verify] {}".format(path))
        integrity.verify(path, hash_path)
        if with_sig:
            sig_path = signature.path_append_sig_suffix(path)
            self.sig_verify(path, sig_path)

    def download_verify_hash(
        self,
        dest_url: str,
        dest_path: Path,
        hash: str,
        *,
        cached: bool = True,
        assume_ok: bool = False
    ) -> None:
        if cached:
            if assume_ok and dest_path.is_file():
                common.vvprint("[assuming OK] {}".format(dest_path))
                return
            try:
                integrity.verify_hash(dest_path, hash)
                common.vprint("[cached file] {}".format(dest_path))
                return
            except (error.MissingFileError, error.IntegrityError):
                pass
        common.vprint("[downloading] {}".format(dest_path))
        self.download(dest_url, dest_path)
        integrity.verify_hash(dest_path, hash)

    def download_verify(
        self,
        dest_url: str,
        dest_path: Path,
        *,
        cached: bool = True,
        assume_ok: bool = False,
        with_sig: bool = False
    ) -> None:
        hash_path = integrity.path_append_hash_suffix(dest_path)
        sig_path = signature.path_append_sig_suffix(dest_path)
        if cached:
            if (
                assume_ok
                and dest_path.is_file()
                and hash_path.is_file()
                and (not with_sig or sig_path.is_file())
            ):
                common.vvprint("[assuming OK] {}".format(dest_path))
                return
            try:
                integrity.verify(dest_path, hash_path)
                if with_sig:
                    self.sig_verify(dest_path, sig_path)
                common.vprint("[cached file] {}".format(dest_path))
                return
            except (error.MissingFileError, error.IntegrityError):
                pass
        common.vprint("[downloading] {}".format(dest_path))
        # Download the (small) hash and signature files first.
        hash_url = integrity.append_hash_suffix(dest_url)
        self.download(hash_url, hash_path)
        if with_sig:
            sig_url = signature.append_sig_suffix(dest_url)
            self.download(sig_url, sig_path)

        # If dest_path exists and has the correct hash, bypass the downloading
        # step to save download time.
        download_required = True
        if dest_path.is_file():
            try:
                integrity.verify(dest_path, hash_path)
                download_required = False
            except (error.MissingFileError, error.IntegrityError):
                pass
        if download_required:
            self.download(dest_url, dest_path)
            integrity.verify(dest_path, hash_path)

        if with_sig:
            self.sig_verify(dest_path, sig_path)



#!/usr/bin/env python3
# coding=utf-8

import argparse
import http.server
import os
from pathlib import Path
from typing import Optional

from romt import common
from romt.crate import crate_prefix_from_name


description = """\
Serve Rust artifacts via http.
"""

epilog = """\
Run a simple HTTP server providing access to mirrored toolchains, rustup, and
crates.  Includes use of git-http-backend to serve the crates index repository.
"""


def add_arguments(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "--port",
        type=int,
        action="store",
        default=8000,
        help="""server listen PORT (default: %(default)s)""",
    )

    parser.add_argument(
        "--bind",
        action="store",
        metavar="ADDR",
        default="localhost",
        help="""change bind ADDR (default: %(default)s)""",
    )


def find_git_cgi_path() -> Optional[Path]:
    cgi_path = Path("cgi-bin")
    if common.is_windows:
        candidates = [
            "git-http-backend.exe",
            "git-http-backend.bat",
        ]
    else:
        candidates = [
            "git-http-backend.sh",
            "git-http-backend",
        ]
    for git_cgi_name in candidates:
        git_path = cgi_path / git_cgi_name
        if git_path.is_file():
            return git_path
    return None


class Handler(http.server.CGIHTTPRequestHandler):
    # Avoid the use of fork on POSIX systems by disabling ``have_fork``.
    # This work-around is required as explained here:
    # https://stackoverflow.com/questions/47758247/ \
    #   local-embedded-http-server-that-can-respond-to-git-commands
    #
    # """
    #   The smart HTTP protocol of git is implemented in git-http-backend. This
    #   is a CGI binary that can be used in combination with some web-servers.
    #   Unfortunately it ignores the content-length of the request and relies
    #   on the server to close its input, which makes it incompatible with some
    #   CGI servers.
    #   ...
    #   [setting ``have_fork`` to False] causes the implementation to use
    #   subprocesses and pipes instead of fork to run the CGI binary. With this
    #   method, the incoming requests are buffered and written to a pipe, which
    #   is then closed. This fits the expected model of git-http-backend and
    #   therefore makes it work.
    # """
    have_fork = False

    def _rewrite_path(self) -> None:
        path = self.path
        if path.startswith("/git/"):
            git_cgi_path = find_git_cgi_path()
            if git_cgi_path is not None:
                path = path.replace("git", git_cgi_path.as_posix(), 1)
        elif path.startswith("/crates/"):
            # /crates/.../<name>/<name>-<version>.crate
            # ->
            # /crates/<prefix>/<name>/<name>-<version>.crate
            parent = os.path.dirname(path)
            name = os.path.basename(parent)
            rel_path = "crates/{}/{}/{}".format(
                crate_prefix_from_name(name), name, os.path.basename(path),
            )
            if os.path.isfile(rel_path):
                path = "/" + rel_path

        if self.path != path:
            common.iprint("Rewrite URL: {} -> {}".format(self.path, path))
            self.path = path

    def do_get(self) -> None:
        self._rewrite_path()
        super().do_GET()

    def do_head(self) -> None:
        self._rewrite_path()
        super().do_HEAD()

    def do_post(self) -> None:
        self._rewrite_path()
        super().do_POST()


# Work-around for non-compliant do_ALLCAPS functions.
setattr(Handler, "do_GET", Handler.do_get)
setattr(Handler, "do_HEAD", Handler.do_head)
setattr(Handler, "do_POST", Handler.do_post)

if common.is_windows:
    GIT_HTTP_BACKEND_SOURCES = [
        "C:/Program Files/Git/mingw64/libexec/git-core/git-http-backend.exe",
    ]
else:
    GIT_HTTP_BACKEND_SOURCES = [
        "/usr/lib/git-core/git-http-backend",
    ]


def get_git_http_backend_path() -> Optional[Path]:
    for source in GIT_HTTP_BACKEND_SOURCES:
        source_path = Path(source)
        if source_path.is_file():
            return source_path
    return None


def make_git_cgi_script(git_http_backend_path: Path) -> None:
    cgi_path = Path("cgi-bin")
    if not cgi_path.is_dir():
        common.iprint("mkdir {}".format(cgi_path))
        cgi_path.mkdir()
    if common.is_windows:
        extension = ".bat"
        template = '@echo off\n"{}"\n'
    else:
        extension = ".sh"
        template = "#!/bin/sh\nexec '{}'\n"
    script_path = cgi_path / ("git-http-backend" + extension)
    script = template.format(git_http_backend_path)
    common.iprint("Create script {} ({})".format(script_path, repr(script)))
    script_path.write_text(script, encoding="utf-8")
    if not common.is_windows:
        common.chmod_executable(script_path)


def setup_git_cgi() -> None:
    os.environ["GIT_HTTP_EXPORT_ALL"] = ""
    os.environ["GIT_PROJECT_ROOT"] = os.path.abspath("git")

    git_cgi_path = find_git_cgi_path()
    if git_cgi_path:
        if not common.is_windows and not common.is_executable(git_cgi_path):
            common.eprint(
                "Warning: setting executable flag on {}".format(git_cgi_path)
            )
            common.chmod_executable(git_cgi_path)
    else:
        git_http_backend_path = get_git_http_backend_path()
        if git_http_backend_path:
            make_git_cgi_script(git_http_backend_path)
        else:
            common.eprint("Warning: missing git-http-backend; no Git support")


class Main:
    def __init__(self, args: argparse.Namespace) -> None:
        self.args = args

    def run(self) -> None:
        common.iprint("server running")
        setup_git_cgi()

        # Avoid type hint warning about missing http.server.test by
        # using getattr() to fetch the function.
        serve_func = getattr(http.server, "test")

        serve_func(
            HandlerClass=Handler, bind=self.args.bind, port=self.args.port
        )



#!/usr/bin/env python3
# coding=utf-8

import copy
from pathlib import Path
from typing import (
    Any,
    Generator,
    Iterable,
    List,
    MutableMapping,
    Optional,
)

import toml

from romt import error


def target_matches_any(target: str, expected_targets: Iterable[str]) -> bool:
    if target == "*":
        return True
    for expected in expected_targets:
        if target == expected or expected == "*":
            return True
    return False


class Package:
    def __init__(
        self, name: str, target: str, details: MutableMapping[str, Any]
    ):
        self.name = name
        self.target = target
        self.available = details["available"]
        self.xz_url = details.get("xz_url", "")

    @property
    def has_rel_path(self) -> bool:
        return self.xz_url != ""

    @property
    def rel_path(self) -> str:
        if not self.has_rel_path:
            raise ValueError(
                "Package {}/{} missing xz_url".format(self.name, self.target)
            )
        url = self.xz_url
        prefix = "/dist/"
        return url[url.index(prefix) + len(prefix) :]


class Manifest:
    def __init__(self, raw_dict: MutableMapping[str, Any]):
        self._dict = raw_dict

    @staticmethod
    def from_toml_path(toml_path: Path) -> "Manifest":
        return Manifest(toml.load(toml_path))

    def clone(self) -> "Manifest":
        return Manifest(copy.deepcopy(self._dict))

    @property
    def _rust_src_version(self) -> str:
        version = self._dict["pkg"]["rust-src"]["version"]
        # Sample version lines found below [pkg.rust-src]:
        # version = "1.43.0-beta.5 (934ae7739 2020-04-06)"
        # version = "1.44.0-nightly (42abbd887 2020-04-07)"
        # version = "1.42.0 (b8cedc004 2020-03-09)"
        return version

    @property
    def channel(self) -> str:
        version = self._rust_src_version
        if "-beta" in version:
            channel = "beta"
        elif "-nightly" in version:
            channel = "nightly"
        else:
            channel = "stable"
        return channel

    @property
    def version(self) -> str:
        version = self._rust_src_version
        # version = "1.44.0-nightly (42abbd887 2020-04-07)"
        # version = "1.42.0 (b8cedc004 2020-03-09)"
        return version.split("-")[0].split()[0]

    @property
    def date(self) -> str:
        return self._dict["date"]

    @property
    def spec(self) -> str:
        return "{}-{}".format(self.channel, self.date)

    @property
    def ident(self) -> str:
        return "{}({})".format(self.spec, self.version)

    def set_package_available(
        self, package_name: str, target: str, available: bool = True
    ) -> None:
        details = self._dict["pkg"][package_name]["target"][target]
        if available and "xz_url" not in details:
            raise error.AbortError(
                "package {}/{} set available but missing xz_url".format(
                    package_name, target
                )
            )
        details["available"] = available

    def get_package(self, package_name: str, target: str) -> Package:
        details = self._dict["pkg"][package_name]["target"][target]
        return Package(package_name, target, details)

    def gen_packages(self) -> Generator[Package, None, None]:
        """Generate Package for all (name, target) in manifest."""
        for name, package_dict in self._dict["pkg"].items():
            for target in package_dict["target"].keys():
                yield self.get_package(name, target)

    def gen_available_packages(
        self, *, targets: Optional[Iterable[str]] = None
    ) -> Generator[Package, None, None]:
        """gen_packages() for available packages matching targets."""
        for package in self.gen_packages():
            if package.available:
                if targets is None or target_matches_any(
                    package.target, targets
                ):
                    yield package

    def available_packages(self) -> List[Package]:
        return list(self.gen_available_packages())

    def _targets_from_packages(self, packages: Iterable[Package]) -> List[str]:
        targets = set(p.target for p in packages)
        targets.discard("*")
        return sorted(targets)

    def all_targets(self) -> List[str]:
        return self._targets_from_packages(self.gen_packages())

    def available_targets(self) -> List[str]:
        return self._targets_from_packages(self.gen_available_packages())
